/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;

import org.simantics.databoard.bindingscheme.DataTypeConstructionException;
import org.simantics.databoard.bindingscheme.RuntimeDataTypeConstructionException;
import org.simantics.databoard.parser.repository.DataTypeRepository;
import org.simantics.databoard.parser.repository.DataTypeSyntaxError;
import org.simantics.databoard.reflection.ReflectionBinding;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.util.StreamUtil;

/**
 * This class is a facade to the data type services. 
 * 
 * @author Toni Kalajainen
 * @author Hannu Niemisto
 */
public class DataTypes {
	
	public static final DataTypeRepository datatypeRepository = new DataTypeRepository();
	
	/**
	 * Read representation from a class. DataType details and parameters
	 * are read as annotations placed in the class.  
	 * (See org.simantics.databoard.annotations)  
	 * <p>
	 * As an exception, in the subclasses of {@link Throwable}, the fields of
	 * Throwable are omited.
	 * 
	 * @see ReflectionBinding  
	 * @param clazz
	 * @return data type
	 * @throws DataTypeConstructionException 
	 */
	public static DataType getDataType(Class<?> clazz) 
	throws DataTypeConstructionException {
		return Bindings.reflectionBinding.getDataType(clazz);
	}
	
	/**
	 * Read representation from a class. DataType details and parameters
	 * are read as annotations placed in the class.  
	 * (See org.simantics.databoard.annotations)  
	 * <p>
	 * This method is used when the caller is 100% sure that binding will be 
	 * constructed without exceptions. Such classes
	 * are all primitive types (Double, Integer, etc, arrays, DataType, ...)
	 * 
	 * This method is unchecked if binding construction to the clazz cannot be trusted.
	 * If construction fails, a RuntimeException is thrown.
	 * 
	 * @see ReflectionBinding  
	 * @param clazz
	 * @return data type
	 * @throws RuntimeDataTypeConstructionException 
	 */
	public static DataType getDataTypeUnchecked(Class<?> clazz) 
	throws RuntimeDataTypeConstructionException {
		try {
			return Bindings.reflectionBinding.getDataType(clazz);
		} catch (DataTypeConstructionException e) {
			throw new RuntimeDataTypeConstructionException(e);
		}
	}	
	
	/**
	 * Adds a type to the repository.
	 * 
	 * @param name Name of the type
	 * @param type Type to be added
	 */
	public static void addDataType(String name, DataType type) {
		datatypeRepository.add(name, type);
	}
	
	/**
	 * Get data type by name.
	 * 
	 * e.g. get("Vec2");
	 * 
	 * @param name
	 * @return data type
	 */
	public static DataType getDataType(String name) {
		return datatypeRepository.get(name);
	}	
	
	/**
	 * Parses and adds type definitions to the repository.
	 * 
	 * The data type can be acquired with #getType
	 * 
	 * e.g. "type Vec2 = { x : Double, y : Double }"
	 * 
	 * @param definitions Definitions in textual format.
	 * @throws DataTypeSyntaxError 
	 */
	public static void addDefinitions(String definitions) throws DataTypeSyntaxError {
		datatypeRepository.addDefinitions(definitions);
	}
	
	/**
	 * Parses an unnamed data type.
	 * 
	 * <a href="https://www.simantics.org/wiki/index.php/Data_type_notation">Datatype Notation</a>
	 * 
	 * e.g. "{ direction : Vec, vector : Vec2 }" 
	 * 
	 * @param typeString The textual representation of the type to be translated
	 * @return Translated data type
	 * @throws DataTypeSyntaxError 
	 */
	public static DataType translate(String typeString) throws DataTypeSyntaxError {
		return datatypeRepository.translate(typeString);
	}	

	static {
		Charset UTF8        = Charset.forName("UTF-8");
		
		// Read File
		InputStream is = DataTypes.class.getResourceAsStream("standardTypes.dbt");
		try {
			String defs = StreamUtil.readString(is, UTF8);
			DataTypes.datatypeRepository.addDefinitions(defs);			
		} catch (IOException e) {
			throw new RuntimeException( e );
		} catch (DataTypeSyntaxError e) {
			throw new RuntimeException( e );
		} finally {
			try { is.close(); } catch (IOException e) {}
		}
		
	}
	
}


