/*******************************************************************************
 * Copyright (c) 2010- Association for Decentralized Information Management in
 * Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;

import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.adapter.Adapter;
import org.simantics.databoard.adapter.AdapterConstructionException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingException;
import org.simantics.databoard.binding.mutable.MutableVariant;
import org.simantics.databoard.binding.mutable.MutableVariantBinding;
import org.simantics.databoard.parser.repository.DataTypeSyntaxError;
import org.simantics.databoard.parser.repository.DataValueRepository;
import org.simantics.databoard.serialization.SerializationException;
import org.simantics.databoard.serialization.Serializer;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.util.StreamUtil;
import org.simantics.databoard.util.binary.BinaryFile;
import org.simantics.databoard.util.binary.BinaryReadable;
import org.simantics.databoard.util.binary.InputStreamReadable;
import org.simantics.databoard.util.binary.RuntimeSerializerConstructionException;

public class Files {

    static final Charset UTF8        = Charset.forName("UTF-8");
	
	/**
	 * Create a text file (.dbv) with a value. The file is UTF-8 encoded.  
	 * If old file exists, it is deleted.
	 * 
	 * @param file
	 * @param binding
	 * @param value
	 * @throws IOException 
	 * @throws BindingException 
	 */
	public static void createTextFile(File file, Binding binding, Object value) 
	throws IOException, BindingException 
	{
		if (file.exists()) file.delete();
		file.createNewFile();
		
		String txt = binding.printValueDefinition(value, false);
		FileOutputStream fos = new FileOutputStream(file, false);
		try {
			OutputStreamWriter os = new OutputStreamWriter(fos, UTF8);
			os.append(txt);
			os.flush();
			os.close();
		} finally {
			fos.close();
		}
	}
	
	/**
	 * Create a binary file (.dbb) with a initial value. Binary file is a variant, 
	 * there is a filetype in the header of the file.
	 * If old file exists, it is deleted.
	 * 
	 * @param file
	 * @param binding
	 * @param value
	 * @throws IOException 
	 * @throws BindingException 
	 * @throws SerializationException 
	 */
	public static void createFile(File file, Binding binding, Object value) 
	throws IOException, SerializationException, BindingException 
	{
		if (file.exists()) file.delete();
		file.createNewFile();
		MutableVariant v = new MutableVariant(binding, value);
		Serializer s = MutableVariantBinding.INSTANCE.serializer();
		s.serialize(v, file);
	}

	/**
	 * Create a new binary file (.dbb) with an empty value. Binary file is a variant, 
	 * there is a filetype in the header of the file.
	 * If old file exists, it is deleted.
	 * 
	 * @since 0.5
	 * @param file
	 * @throws IOException 
	 * @throws BindingException 
	 * @throws SerializationException 
	 * @throws RuntimeSerializerConstructionException
	 */
	public static void createFile(File file) throws IOException, RuntimeSerializerConstructionException, SerializationException, BindingException  
	{
		if (file.exists()) file.delete();
		file.createNewFile();
		BinaryFile bf = new BinaryFile(file);
		try {
			Binding binding = MutableVariantBinding.INSTANCE;
			Object value = new MutableVariant(); 
			binding.serializer().serialize(bf, value);
		} finally {
			bf.close();
		}
	}	

	/**
	 * Create a binary file (.dbb) with empty value of given type. Binary file is a variant, 
	 * there is a filetype in the header of the file.
	 * If old file exists, it is deleted.
	 * 
	 * @param file
	 * @param type
	 * @throws IOException 
	 * @throws BindingException 
	 * @throws SerializationException 
	 * @throws RuntimeSerializerConstructionException 
	 */
	public static void createFile(File file, DataType type) throws IOException, BindingException, RuntimeSerializerConstructionException, SerializationException 
	{
		if (file.exists()) file.delete();
		file.createNewFile();
		BinaryFile bf = new BinaryFile(file);
		try {
			Binding binding = MutableVariantBinding.INSTANCE;
			Object value = binding.createDefault(); 
			binding.serializer().serialize(bf, value);
		} finally {
			bf.close();
		}
	}
	
	/**
	 * Read a text file (.dbv).   
	 * 
	 * @param file
	 * @param binding
	 * @return value
	 * @throws IOException 
	 * @throws BindingException 
	 * @throws DataTypeSyntaxError 
	 */
	public static Object readTextFile(File file, Binding binding) 
	throws IOException, DataTypeSyntaxError, BindingException 
	{
		if (file.exists()) file.delete();
		file.createNewFile();
		FileInputStream fis = new FileInputStream(file);
		try {
			byte[] data = StreamUtil.readFully(fis);
			String txt = new String(data, UTF8);
			DataValueRepository repo = new DataValueRepository();
			repo.setTypeRepository( DataTypes.datatypeRepository );
			return binding.parseValue(txt, repo);
		} finally {
			fis.close();
		}
	}
	
	/**
	 * Read file type of a binary file. 
	 * 
	 * @param file
	 * @return
	 * @throws IOException 
	 * @throws BindingException 
	 * @throws SerializationException 
	 */
	public static DataType readFileType(File file) throws IOException, SerializationException, BindingException {
		BinaryFile rf = new BinaryFile( file );
		try {
			Binding datatype_binding = Bindings.getBindingUnchecked( DataType.class );
			return (DataType) datatype_binding.serializer().deserialize( rf );
		} finally {
			rf.close();
		}		
	}
	
	/**
	 * Read a binary file into a java instance. Binary file is a variant, 
	 * there is a filetype in the header of the file. 
	 * If requested binding is not the exact binding of the file, an adapter is tried.
	 * 
	 * @param file file
	 * @param binding content binding
	 * @return instance
	 * @throws SerializationException 
	 * @throws AccessorException 
	 */
	public static Object readFile(File file, Binding binding) throws IOException, SerializationException {
		BinaryFile rf = new BinaryFile( file );
		try {
			Binding datatype_binding = Bindings.getBindingUnchecked( DataType.class );
			DataType type = (DataType) datatype_binding.serializer().deserialize( rf );
			
			if (type.equals(binding.type())) {
				return binding.serializer().deserialize(rf);
			} else {
				try {
					Binding fileContentBinding = Bindings.getBinding(type);
					Adapter adapter = Bindings.getAdapter(fileContentBinding, binding);
					Object value = fileContentBinding.serializer().deserialize(rf);
					return adapter.adapt( value );
				} catch (AdapterConstructionException e) {
					throw new SerializationException(e);
				} catch (AdaptException e) {
					throw new SerializationException(e);
				}
			}
			
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		} finally {
			rf.close();
		}
	}

	/**
	 * Read input stream into a java instance. Binary file is a variant, 
	 * there is a filetype in the header of the file. If requested binding is not the 
	 * exact binding of the file, an adapter is tried.<p>
	 * 
	 * The implementation reads the inputstream fully into memory.<p>
	 * 
	 * @param is input stream
	 * @param binding content binding
	 * @return instance
	 * @throws SerializationException 
	 * @throws AccessorException 
	 */
	public static Object readFile(InputStream is, Binding binding) throws IOException, SerializationException {
		try {
			BinaryReadable readable = InputStreamReadable.readFully( is );
			Binding datatype_binding = Bindings.getBindingUnchecked( DataType.class );
			DataType type = (DataType) datatype_binding.serializer().deserialize( readable );
			
			if (!type.equals(binding.type())) {
				try {
					Binding fileContentBinding = Bindings.getBinding(type);
					Adapter adapter = Bindings.getAdapter(fileContentBinding, binding);
					Object value = fileContentBinding.serializer().deserialize( readable );
					return adapter.adapt( value );
				} catch (AdapterConstructionException e) {
					throw new SerializationException(e);
				} catch (AdaptException e) {
					throw new SerializationException(e);
				}
			}
			
			return binding.serializer().deserialize( readable );
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	/**
	 * Read input stream into a java instance. If requested binding is not the 
	 * exact binding of the file, an adapter is tried.
	 * 
	 * @param is input stream
	 * @param streamLength
	 * @param binding content binding
	 * @return instance
	 * @throws SerializationException 
	 * @throws AccessorException 
	 */
	public static Object readFile(InputStream is, long streamLength, Binding binding) throws IOException, SerializationException {
		try {
			BinaryReadable readable = new InputStreamReadable( is, streamLength );
			Binding datatype_binding = Bindings.getBindingUnchecked( DataType.class );
			DataType type = (DataType) datatype_binding.serializer().deserialize( readable );
			
			if (!type.equals(binding.type())) {
				try {
					Binding fileContentBinding = Bindings.getBinding(type);
					Adapter adapter = Bindings.getAdapter(fileContentBinding, binding);
					Object value = fileContentBinding.serializer().deserialize( readable );
					return adapter.adapt( value );
				} catch (AdapterConstructionException e) {
					throw new SerializationException(e);
				} catch (AdaptException e) {
					throw new SerializationException(e);
				}
			}
			
			return binding.serializer().deserialize( readable );
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
}

