/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.binding;

import gnu.trove.TObjectIntHashMap;

import java.util.IdentityHashMap;
import java.util.Set;

import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingException;
import org.simantics.databoard.binding.mutable.MutableInteger;
import org.simantics.databoard.type.OptionalType;
import org.simantics.databoard.util.IdentityPair;


/**
 * This is a binding of Optional Type and a Java Object.
 * 
 * @see OptionalType
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public abstract class OptionalBinding extends Binding {
	
    public Binding componentBinding;

    public OptionalBinding(OptionalType type, Binding componentBinding) {
        this.componentBinding = componentBinding;
        this.type = type;        
    }
    
    public OptionalBinding(Binding componentBinding) {
        this.componentBinding = componentBinding;
        this.type = new OptionalType(componentBinding.type());
    }
    
    @Override
    public OptionalType type() {
    	return (OptionalType) type;
    }
    
    /**
     * Create result with no value
     * 
     * @return no value
     */
    public abstract Object createNoValue() throws BindingException;
        
    /**
     * Create result with a value
     * 
     * @param value
     * @return argument that contains a value
     */
    public abstract Object createValue(Object value) throws BindingException;


    /**
     * Tests whether arg contains a value
     * 
     * @param arg 
     * @return true if arg contained a value
     */
    public abstract boolean hasValue(Object arg) throws BindingException;
    
    /**
     * Get the non-null value, the arg did not contain a value,
     * BindingException is thrown. 
     *  
     * @param arg argument that contains a value
     * @return the composite value
     * @throws BindingException
     */
    public abstract Object getValue(Object arg) throws BindingException;

    public abstract void setValue(Object optional, Object componentValue) throws BindingException;
    
    public abstract void setNoValue(Object optional) throws BindingException;
    
	public abstract boolean isInstance(Object obj);
    
    public Binding getComponentBinding() {
        return componentBinding;
    }
    
    @Override
    public void accept(Visitor1 v, Object obj) {
        v.visit(this, obj);        
    }
    
    @Override
    public <T> T accept(Visitor<T> v) {
        return v.visit(this);
    }
	
	/**
	 * Assert tje obj is a correct instance.
	 * Assert the obj is valid also according to the component binding.
	 */
	@Override
	public void assertInstaceIsValid(Object obj, Set<Object> validInstances) throws BindingException {
		if (!hasValue(obj)) return;
		Object componentValue = getValue(obj);
		componentBinding.assertInstaceIsValid(componentValue, validInstances);
	}
	
	@Override
    public int deepHashValue(Object value, IdentityHashMap<Object, Object> hashedObjects) throws BindingException {
		return hasValue(value) ? componentBinding.deepHashValue( getValue(value), hashedObjects ) : 0;
	}
	
	@Override
	public int deepCompare(Object o1, Object o2,
			Set<IdentityPair<Object, Object>> compareHistory)
			throws BindingException {
		Boolean h1 = hasValue(o1);
		Boolean h2 = hasValue(o2);
		if (!h1 && !h2) return 0;		
		int dif = h1.compareTo(h2);
		if (dif!=0) return dif;
		
		Binding c = getComponentBinding();
		Object v1 = getValue(o1);
		Object v2 = getValue(o2);
		return c.deepCompare(v1, v2, compareHistory);		
	}
	
    public Object createNoValueUnchecked() throws RuntimeBindingException {
    	try {
    		return createNoValue();
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
    }
    
    public Object createValueUnchecked(Object value) 
    throws RuntimeBindingException {
    	try {
			return createValue(value);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
    }
   
    public boolean hasValueUnchecked(Object arg) throws RuntimeBindingException {
    	try {
			return hasValue(arg);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
    }

    @Override
    protected void toString(Object value, StringBuilder b, TObjectIntHashMap<Object> refs, MutableInteger nameCount) throws BindingException {
    	if(hasValue(value))
    		getComponentBinding().toString(getValue(value), b, refs, nameCount);
    	else
    		b.append("null");
    }
    
}

