/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.binding;

import gnu.trove.TObjectIntHashMap;

import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Set;

import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingException;
import org.simantics.databoard.binding.mutable.MutableInteger;
import org.simantics.databoard.type.Component;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.util.IdentityHashSet;
import org.simantics.databoard.util.IdentityPair;


/**
 * This is a binding of a Record Type and a Java Object.
 * 
 * @see RecordType
 * @author Hannu Niemisto
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public abstract class RecordBinding extends Binding {
    
    public Binding[] componentBindings;
    
    /**
     * Get binding by field name
     * @param fieldName
     * @return binding or <code>null</code>
     */
    public Binding getComponentBinding(String fieldName) {
    	int fieldIndex = type().getComponentIndex2(fieldName);
    	if (fieldIndex<0) return null;
    	return getComponentBinding(fieldIndex);
    }
    
    public Binding getComponentBinding(int fieldIndex) {
    	return componentBindings[fieldIndex];
    }
    
	public Binding[] getComponentBindings() {
	    return componentBindings;
	}
	
	protected void setComponentBindings(Binding[] componentBindings) {
		this.componentBindings = componentBindings;
	}
	
	@Override
	public RecordType type() {
		return (RecordType) type;
	}
	
	public int count() {
		return type().count();
	}
	
	public abstract Object getComponent(Object obj, int index)
	throws BindingException;
	
	/**
	 * Create a record using values.
	 * 
	 * Note! values may be consumed (used in the result)
	 * 
	 * @param values
	 * @return new record
	 * @throws BindingException
	 */
	public abstract Object create(Object ... values)
	throws BindingException;
	
	/**
	 * Creates partial and most likely invalid instance.
	 * This is used in two-phase construction of recursive instances.
	 *  
	 * @return instance.
	 */
	public abstract Object createPartial() throws BindingException;
	
	public abstract void setComponents(Object obj, Object ... value) throws BindingException;
	public abstract void setComponent(Object obj, int index, Object value) throws BindingException;
	
    @Override
    public void accept(Visitor1 v, Object obj) {
        v.visit(this, obj);        
    }
    
    @Override
    public <T> T accept(Visitor<T> v) {
        return v.visit(this);
    }
    
    /**
     * Assert obj is valid Record Type
     * 
     * This asserts all fields are valid. 
     * 
     * @throws BindingException
     */
    @Override
    public void assertInstaceIsValid(Object obj, Set<Object> validInstances) throws BindingException {
    	if (!isInstance(obj)) 
    		throw new BindingException("The object ("+obj+") is not correct instance.");
    	RecordType type = type();    	
    	int length = type.count();
    	
    	if (type.isReferable()) {
    		if (validInstances==null) {
    			validInstances = new IdentityHashSet<Object>();
    		}
    		if (validInstances.contains(obj)) return;
    		validInstances.add(obj);
    	}
    	
    	for (int i=0; i<length; i++)
    	{
    		Binding componentBinding = getComponentBindings()[i]; 
    		Object componentValue = getComponent(obj, i);
    		componentBinding.assertInstaceIsValid(componentValue, validInstances);
    	}
    }
    
    @Override
    public int deepHashValue(Object value, IdentityHashMap<Object, Object> hashedObjects) throws BindingException {
    	if (type().isReferable()) {
	    	if (hashedObjects==null) {
	    		hashedObjects = new IdentityHashMap<Object, Object>(1);
	    	}
	    	if (hashedObjects.containsKey(value)) {
	    		return 0;    	
	    	} else {
	    		hashedObjects.put(value, null);
	    	}
    	}
    	
    	int result = 3;
    	int len = componentBindings.length;
    	for (int i=0; i<len; i++) {
    		Object element = getComponent(value, i);
    		result += 31*result + componentBindings[i].deepHashValue(element, hashedObjects);
    	}
    	return result;
    }
    
    @Override
    public int deepCompare(Object o1, Object o2,
    		Set<IdentityPair<Object, Object>> compareHistory)
    		throws BindingException {
		RecordType rt    = (RecordType) type(); 
		
		// Compare History is used to prevent infinite loops
		if (rt.isReferable()) {
        	if (compareHistory==null) {
        		compareHistory = new HashSet<IdentityPair<Object, Object>>(); 
        	} 	
        	IdentityPair<Object, Object> p = new IdentityPair<Object, Object>(o1, o2);
        	if (compareHistory.contains(p)) return 0;            	
    		compareHistory.add(p);
		}
		
 		int len = rt.count();
		for (int i=0; i<len; i++) {
			Binding c = getComponentBindings()[i];
			Object v1 = getComponent(o1, i);
			Object v2 = getComponent(o2, i);
			int dif = c.deepCompare(v1, v2, compareHistory);
			if (dif!=0) return dif;
		}
		return 0;
    }
	
	public Object createUnchecked(Object ... values) throws RuntimeBindingException
	{
		try {
			return create(values);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	public void setComponentsUnchecked(Object obj, Object ... value) throws RuntimeBindingException {
		try {
			setComponents(obj, value);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	public void setComponentUnchecked(Object obj, int index, Object value) throws RuntimeBindingException {
		try {
			setComponent(obj, index, value);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}

	private void toStringAux(Object value, StringBuilder b, TObjectIntHashMap<Object> refs, MutableInteger nameCount) throws BindingException {
		b.append('{');
		Component[] components = type().getComponents();
		boolean first = true;
		for(int i=0;i<components.length;++i) {
			if(first)
				first = false;
			else
				b.append(", ");
			Binding binding = getComponentBinding(i);
			Object cval = getComponent(value, i);
			if(binding instanceof OptionalBinding && 
					!((OptionalBinding)binding).hasValue(cval))
				continue;
			Component component = components[i];			
			b.append(component.name);
			b.append(" = ");
			binding.toString(cval, b, refs, nameCount);
		}
		b.append('}');
	}
	
    @Override
    protected void toString(Object value, StringBuilder b, TObjectIntHashMap<Object> refs, MutableInteger nameCount) throws BindingException {
    	if(type().referable) {
    		if(refs.contains(value)) {
    			int val = refs.get(value);
    			if(val < 0) {
    				val = nameCount.value++;
    				refs.put(value, val);
    			}
    			b.append((char)('a' + val));
    		}
    		else {
    			refs.put(value, -1);
    			toStringAux(value, b, refs, nameCount);
    			int val = refs.remove(value);
    			if(val >= 0) {
    				b.append("/");
    				b.append((char)('a' + val));
    			}
    		}
    	}
    	else {
    		toStringAux(value, b, refs, nameCount);
    	}
	}
	
}

